/*
 * inv.cc - a C++ class for the data determining GW invariants
 *
 */

#include <stdlib.h>
#include <iostream>
#include <string.h>
#include <assert.h>

#include "inv.h"
#include "config.h"



// Compare two pdatas. It is crucial for many other functions that mult is
// the most significant sorting criterion, followed by psipow and finally
// Hpow.

int cmp (const pdata& x, const pdata& y)
{
	if (x.mult<y.mult) return -1;
	if (x.mult>y.mult) return 1;

	if (x.psipow<y.psipow) return -1;
	if (x.psipow>y.psipow) return 1;

	if (x.Hpow<y.Hpow) return -1;
	if (x.Hpow>y.Hpow) return 1;

	return 0;
}



// Allocate memory for points

void inv::realloc (int n) {
	if ( n>alloc ) {
		pdata *p = pt;
		assert (pt = new pdata [n]);
		if (points) memmove (pt,p,points*sizeof(pdata));
		delete [] p;
		alloc = n;
	}
}



// Sort points according to their classes. Sort only the points from
// low to high. A negative argument means the beginning (resp. end)
// of the list.

void inv::internal_sort (int low, int high)
{
	if (low<0) low = 0;
	if (high<0) high = points-1;

	if (low>=high) return;

	int lo = low;
	int hi = high+1;

	for (;;) {
		while (++lo<points && pt[lo] < pt[low]);
		while (--hi>=0 && pt[hi] > pt[low]);
		if (lo<hi) swap (lo,hi);
		else break;
	}
	swap (low,hi);

	if (hi>1) internal_sort (low,hi-1);
	internal_sort (hi+1,high);
}



/*
 * Reduce the primitive classes to the smallest possible. It's best to
 * apply this to a sorted class. The result need not be sorted in the strict
 * sense, but it is guaranteed that a call to sort(), followed by a call to
 * reduce() will reduce in the same invariant whenever this is possible.
 * The only point where it may happen that the invariant is not actually sorted
 * is that primitive classes with the same multiplicity and psi power are in
 * the wrong order. Return the smallest primitive class that occurs in the
 * invariant. The function terminates very quickly if the invariant has not
 * changed since the last reduce() call.
 *
 */

int inv::reduce ()
{
	int i,j,k,l;

	j = 0; // primitive classes -1,-2,...,j have already been assigned

	for (i=0; i<points; i++) if ( (k=pt[i].Hpow)<j && k != --j ) {
		// Exchange primitive classes k and j
		for (l=i; l<points; l++) {
			if (pt[l].Hpow==k) pt[l].Hpow=j;
			else if (pt[l].Hpow==j) pt[l].Hpow=k;
		}
	}

	return j;
}



// Delete a marked point

void inv::del (int n)
{
	points--;
	if (n<points) memmove (&pt[n],&pt[n+1],(points-n)*sizeof(pdata));
}



// Output (mainly for debugging)

ostream& operator<< (ostream& os, const pdata& x)
{
	if (x.mult) os << '[';
	if (x.Hpow>0) {
		os << 'H';
		if (x.Hpow>1) os << x.Hpow;
	}
	if (x.Hpow<0) {
		os << "pc(" << -x.Hpow << ")";
	}
	if (x.psipow) {
		os << 'c';
		if (x.psipow>1) os << x.psipow;
	}
	if (! x.Hpow && ! x.psipow ) os << '1';
	if (x.mult) os << ',' << x.mult << ']';

	return os;
}



ostream& operator<< (ostream& os, const inv& x)
{
	int n;
	os << "<n=" << x.xdim << ",e=" << x.xdeg << ",g=" << x.cgenus
		<< ",d=" << x.cdeg;
	for (int i=0; i<x.points; i++) {
		os << ',' << x.pt[i];
		n=1;
		while (i<x.points-1 && x.pt[i] == x.pt[i+1]) {
			i++; n++;
		}
		if (n>1) os << ':' << n;
	}
	os << ">";
	return os;
}



#ifdef INV_DEBUG

// Check the invariant for invalid entries

void inv::check () const
{
	int i;
	int vd = 0; // the virtual dimension
	int rel = 0; // do we have a relative invariant?
	int tm = 0; // the total multiplicity

	// Check for entries out of range
	if (xdim<0) cout << "*inv* " << *this
			 << " wrong dimension N=" << xdim << endl;
	if (xdeg<1) cout << "*inv* " << *this
			 << " wrong degree E=" << xdeg << endl;
	if (cdeg<0 || (cdeg && !xdim))
		cout << "*inv* " << *this
		     << " wrong degree D=" << cdeg << endl;
	if (points<0) cout << "*inv* wrong number of points " << points <<endl;
	for (i=0; i<points; i++) {
		if (pt[i].psipow<0)
			cout << "*inv* " << *this
			     << " wrong psi power " << pt[i].psipow<<endl;
		if (pt[i].mult<0)
			cout << "*inv* " << *this
			     << " wrong multiplicity " <<pt[i].mult<<endl;
		if (pt[i].Hpow<0 && xdeg==1)
			cout << "*inv* " << *this
			     << " no primitive classes in P^n" << endl;
		if (pt[i].mult) rel = 1; // a relative invariant
		vd -= pt[i].vdim (xdim);
		tm += pt[i].mult;
	}

	for (i=0; i<points; i++)
		if (pt[i].Hpow>xdim+1 || (pt[i].Hpow>xdim &&
					(!rel || pt[i].mult)))
			cout << "*inv* " << *this
			     << " wrong H power " << pt[i].Hpow << endl;

	// Check the virtual dimension. Note that if the invariant is relative,
	// it's in fact an invariant of the ambient P^(n+1).
	if (rel) {
		if (cgenus<0) {
			cout << "*inv* " << *this
			     << " relative negative genus invariant" << endl;
		}
		vd += 2*((2+xdim)*cdeg + (xdim-2)*(1-cgenus));
	} else {
		if (cgenus<0) {
			vd += 2*((2+xdim-xdeg)*cdeg + xdim-3+cgenus);
		} else {
			vd += 2*((2+xdim-xdeg)*cdeg + (xdim-3)*(1-cgenus));
		}
	}

	if (vd) cout << "*inv* " << *this
		     << " virtual dimension is " << vd << endl;

	// A negative genus invariant must have at least 2 marked points,
	// the second one can have no psi classes
	if (cgenus<0) {
		if (points<2) {
			cout << "*inv* " << *this
			     << " must have at least 2 marked points" << endl;
		}
		if (pt[1].psipow) {
			cout << "*inv* " << *this
			     << " psi class on second point" << endl;
		}
	}

	// Check the total multiplicity
	if (tm>xdeg*cdeg) cout << "*inv* " << *this
			       << " total multiplicity is " << tm << endl;
}

#endif



// Database functions. We store the invariants in the obvious way, with
// the exception that a negative number -n indicates that the next point
// occurs n times.

void set_key (database& db, const inv& x)
{
	int n;

	set_key (db, x.xdim);
	set_key (db, x.xdeg);
	set_key (db, x.cgenus);
	set_key (db, x.cdeg);
	set_key (db, x.points);
	for (int i=0; i<x.points; i++) {
		int n = -1;
		while (i<x.points-1 && x.pt[i] == x.pt[i+1]) {
			i++; n--;
		}
		if (n<-1) set_key (db, n);
		set_key (db, x.pt[i].mult);
		set_key (db, x.pt[i].psipow);
		set_key (db, x.pt[i].Hpow);
	}
}



void set_value (database& db, const inv& x)
{
	int n;
	set_value (db, x.xdim);
	set_value (db, x.xdeg);
	set_value (db, x.cgenus);
	set_value (db, x.cdeg);
	set_value (db, x.points);
	for (int i=0; i<x.points; i++) {
		int n = -1;
		while (i<x.points-1 && x.pt[i] == x.pt[i+1]) {
			i++; n--;
		}
		if (n<-1) set_key (db, n);
		set_value (db, x.pt[i].mult);
		set_value (db, x.pt[i].psipow);
		set_value (db, x.pt[i].Hpow);
	}
}



bool get_value (database& db, inv& x)
{
	int n;

	if ( ! get_value (db, x.xdim) ) return 0;
	if ( ! get_value (db, x.xdeg) ) return 0;
	if ( ! get_value (db, x.cgenus) ) return 0;
	if ( ! get_value (db, x.cdeg) ) return 0;
	if ( ! get_value (db, x.points) ) return 0;
	x.realloc (x.points);
	for (int i=0; i<x.points; i++) {
		n = 1;
		if ( ! get_value (db, x.pt[i].mult) ) return 0;
		if ( x.pt[i].mult < 0) {
			n = -x.pt[i].mult;
			if ( ! get_value (db, x.pt[i].mult) ) return 0;
		}
		if ( ! get_value (db, x.pt[i].psipow) ) return 0;
		if ( ! get_value (db, x.pt[i].Hpow) ) return 0;
		for (; n>1; n--) {
			x.pt[i+1] = x.pt[i];
			i++;
		}
	}

	return 1;
}

