/*
 * simplify.cc - simplify GW invariants using the standard axioms
 *
 */

#include <stdlib.h>
#include <iostream>
#include <assert.h>

#include "simplify.h"

/*
 * If we include a file setting SIMPLIFY_DEBUG to a positive integer, debugging
 * output is printed for all computations in this file of invariants up to
 * this level. This shows how the invariants are computed.
 * 
 */

#include "config.h"



#define NN iv.xdim
#define EE iv.xdeg
#define GG iv.cgenus
#define DD iv.cdeg



// Debugging macros

#define DEBUG_BEFORE(TEXT)						\
if (lvl<=SIMPLIFY_DEBUG)						\
	cout << "*simplify* level " << lvl				\
	     << ": trying " TEXT << endl;
#define DEBUG_AFTER							\
if (lvl<=SIMPLIFY_DEBUG)						\
	cout << "*simplify* level " << lvl				\
	     << ": result is " << result << endl;



// Compute the degree-0 invariants. It is necessarily an absolute invariant.

rational mapping_to_point (inv& iv)
{
	rational result;
	int i,j,k,l;

	if (!GG) {
		// genus-0 mapping to point: multinomial coefficient times
		// an ordinary intersection product.
		result = EE; // the top H power gives a factor of EE
		k = 0; // factorial in the numerator
		l = 0; // H power
		for (i=0; i<iv.points; i++) {

			// The H power / primitive class:
			// Zero if H*primitive or primitive*H
			if (iv.pt[i].Hpow>0 && l<0) return 0;
			if (iv.pt[i].Hpow<0) {
				if (!l) l = iv.pt[i].Hpow;
				else if (iv.pt[i].Hpow!=l) return 0;
				else {
					l=iv.xdim;
					result /= EE;
				}
			} else l += iv.pt[i].Hpow;

			// The binomial coefficient
			for (j=1; j<=iv.pt[i].psipow; j++) {
				result *= (++k);
				result /= j;
			}
		}
		if ( l!=NN ) return 0;
		return result;
	}

	if (GG==1) {
		/*
		 * genus-1 mapping to point
		 *
		 * See Kontsevich-Manin for (some) details.
		 *
		 * In genus 1, the dimension condition guarantees that
		 * the fundamental class, cotangent line and divisor axioms
		 * can reduce everything to the 1-pointed case.
		 *
		 * Actually, it is easily seen that we must have a total H
		 * power of either 0 or 1. In the 1-pointed case, the invariant
		 * is then more or less the top or (top-1) Chern class of the
		 * variety. We compute this below (using the standard Euler
		 * and normal bundle sequence).
		 *
		 */
		if (!iv.points) return 0; // the 0-point invariant is 0
		// Compute the total H power; it must be 0 or 1. No primitive
		// classes can occur.
		l = 0;
		for (i=0; i<iv.points; i++) {
			if (iv.pt[i].Hpow<0) return 0;
			l += iv.pt[i].Hpow;
		}
		if (l>=2) return 0;
		if (iv.points>1) return rational ();
		if (l)
			// The invariant <H>. It's equal to minus the (top-1)
			// Chern class of the variety times the integral
			// <lambda> on \bar M_{1,1}, which is 1/24.
			return ( pow(rational(1-EE),NN+2) - 1 -
				 (NN+2)*EE*(rational(EE)/2*(NN+1)-1) )
				/(24*EE*EE);
		// The invariant <psi>. It's equal to the top Chern class of
		// the variety times the <psi> integral over \bar M_{1,1},
		// which is 1/24.
		return (pow(rational(1-EE),NN+2)+(NN+2)*EE-1)/(24*EE);
	}

	// Currently we don't know any other d=0 invariants
	return rational ();
}



// The 1-point rational invariants of P^n. We're allowed to change the
// invariant. The formula is that the J function is
//   prod_{i=1}^d 1/(H+i)^(n+1).
// We do this recursively over d, coding the binomial expansion of
// 1/(H+d)^(n+1) directly.

rational projspace_1point (inv& iw, int lvl)
{
	rational result;
	rational coeff;
	int i,j;

	if (iw.pt[0].Hpow<0) return 0; // No primitive classes!

	if (iw.cdeg==1) {
		// Starting the recursion for degree 1. If h is the H power.
		// the result is binomial (-n-1,n-h).
		result = 1;
		j = -iw.xdim-1;
		for (i=1; i<=iw.xdim-iw.pt[0].Hpow; i++) {
			result *= (j--);
			result /= i;
		}
		return result;
	}

	// The recursion step.
	result = 0;
	j = 0;
	coeff = pow (rational(iw.cdeg),-iw.xdim-1);
	iw.cdeg--;
	iw.pt[0].psipow -= (iw.xdim+1);
	while (iw.pt[0].Hpow<=iw.xdim) {
		result += coeff * compute (iw, lvl+1);
		j++;
		// this is to update the binomial coefficient:
		coeff *= (-iw.xdim-j);
		coeff /= (j*(iw.cdeg+1));
		iw.pt[0].Hpow++;
		iw.pt[0].psipow--;
	}
	return result;
}



// Check for various simplifications. Return UNDEFINED if this doesn't work
// to compute the invariant. In this case, return the invariant in sorted form.
// If the computation succeeds and is not trivial, we store the result in
// the database.

rational simplify (inv& iv, int lvl)
{
	inv iw;
	rational result;
	int i,j;

	// Degree 0 -> try mapping to point axiom. This is guaranteed to
	// succeed for <=3 marked points in genus 0 and <=1 marked point in
	// genus 1. So it's OK after this to apply axioms that reduce the
	// number of marked points by 1.
	if ( (!DD) && (GG>=0)) {
		#ifdef SIMPLIFY_DEBUG
		DEBUG_BEFORE ("mapping to point axiom");
		#endif
		result = mapping_to_point (iv);
		#ifdef SIMPLIFY_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) return result;
	}

	// Sort the invariant.
	iv.sort ();
	#ifdef SIMPLIFY_DEBUG
	if (lvl<=SIMPLIFY_DEBUG)
		cout << "*simplify* level " << lvl
		     << ": sort invariant -> " << iv << endl;
	#endif

	// Reduce the invariant (i.e. normalize primitive classes).
	if (EE!=1) {
		iv.reduce ();
		#ifdef SIMPLIFY_DEBUG
		if (lvl<=SIMPLIFY_DEBUG)
			cout << "*simplify* level " << lvl
			     << ": reduce invariant -> " << iv << endl;
		#endif
	}

	// Look the invariant up in the database.
	#ifdef SIMPLIFY_DEBUG
	DEBUG_BEFORE ("database lookup");
	#endif
	GWdbm.reset_key ();
	set_key (GWdbm, iv);
	if ( GWdbm.read () ) {
		get_value (GWdbm, result);
		#ifdef SIMPLIFY_DEBUG
		DEBUG_AFTER;
		#endif
		return result;
	}
	#ifdef SIMPLIFY_DEBUG
	DEBUG_AFTER;
	#endif

	// For relative invariants we don't know any axioms right now.
	if ( iv.points && iv.pt[iv.points-1].mult ) return rational ();

	// Neither for negative genus invariants.
	if ( GG<0 ) return rational ();

	// Check for the cotangent line axiom in an absolute invariant. This
	// is also very fast...
	// We can skip the loop if we see at the beginning that there are
	// no descendants at all. We can also break the loop after we read
	// a psi power bigger than 1.
	// We do not do this in genus bigger than 1 if we would have to
	// delete the last marked point.
	if ( iv.points && iv.pt[iv.points-1].psipow &&
			(GG<=1 || iv.points>1) )
	for (i=0; i<iv.points; i++) {
		if ( iv.pt[i].psipow>1 ) break;
		if ( iv.pt[i].Hpow || iv.pt[i].psipow!=1 ) continue;

		#ifdef SIMPLIFY_DEBUG
		DEBUG_BEFORE ("cotangent line axiom");
		#endif
		iw = iv;
		iw.del (i);
		if (2*GG-2+iw.points)
			result = (2*GG-2+iw.points)*compute (iw,lvl+1);
		else result = 0;
		#ifdef SIMPLIFY_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) goto STORE;
		break; // If one cotangent line class doesn't
		       // work, the others won't either.
	}

	// Check for the rational 1-point invariants of P^n.
	if ( iv.points==1 && !GG && EE==1 ) {
		#ifdef SIMPLIFY_DEBUG
		DEBUG_BEFORE ("P^n 1-point invariant rule");
		#endif
		iw = iv;
		result = projspace_1point (iw,lvl);
		#ifdef SIMPLIFY_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) goto STORE;
	}

	// Check for the fundamental class axiom. We can skip the loop looking
	// for a fundamental class if we find an entry with positive Hpow or
	// psipow.
	if (GG<=1 || iv.points>1) for (i=0; i<iv.points; i++) {
		if (iv.pt[i].Hpow>0 || iv.pt[i].psipow) break;
		if (iv.pt[i].Hpow) continue; // read a primitive class

		#ifdef SIMPLIFY_DEBUG
		DEBUG_BEFORE ("fundamental class axiom");
		#endif
		// Check first if there are no descendants - in this case the
		// invariant is 0.
		if ( !iv.pt[iv.points-1].psipow ) result = 0;
		else {
			iw = iv;
			iw.del (i);
			result = 0;
			// As the invariant is sorted, we can break the
			// loop after the first non-descendant class
			for (i=iw.points-1; i>=0 && iw.pt[i].psipow; i--) {
				iw.pt[i].psipow--;
				result += compute (iw,lvl+1);
				iw.pt[i].psipow++;
			}
		}
		#ifdef SIMPLIFY_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) goto STORE;
		break; // If one fundamental class doesn't work, the others
		       // won't either.
	}

	// Check for the divisor axiom. Use that the invariant is sorted to
	// exit the loop as soon as possible. Divisor classes that are
	// primitive classes (if the dimension of the variety is 2) are not
	// treated here.
	if (GG>=1 || iv.points>1) for (i=0; i<iv.points; i++) {
		if (iv.pt[i].psipow || iv.pt[i].Hpow>1) break;
		if (!iv.pt[i].Hpow) continue;

		// We've found a divisor class.
		#ifdef SIMPLIFY_DEBUG
		DEBUG_BEFORE ("divisor axiom");
		#endif
		iw = iv;
		iw.del (i);
		if (DD) result = DD*compute (iw,lvl+1); // the standard term
		else result = 0;
		for (j=iw.points-1; j>=0 && iw.pt[j].psipow; j--)
			if (iw.pt[j].Hpow<NN && iw.pt[j].Hpow>=0) {
				iw.pt[j].psipow--;
				iw.pt[j].Hpow++;
				result += compute (iw,lvl+1); // corrections
				iw.pt[j].Hpow--;
				iw.pt[j].psipow++;
			}
		#ifdef SIMPLIFY_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) goto STORE;
		break; // If one divisor class didn't work, the others
		       // won't either.
	}

	// Check for genus>1 invariants with no marked points. To be able to
	// compute these with the Virasoro conditions we have to add a
	// marked point.
	if ( (!iv.points) && GG>1 ) {
		#ifdef SIMPLIFY_DEBUG
		DEBUG_BEFORE ("point addition");
		#endif
		iv.realloc(1);
		iv.points = 1;
		iv.pt[0] = pdata (0,1);
		result = compute (iv,lvl+1)/(2*iv.cgenus-2);
		iv.points = 0;
		#ifdef SIMPLIFY_DEBUG
		DEBUG_AFTER;
		#endif
		if ( ! is_undefined (result) ) goto STORE;
	}

	// Sorry, but we can't simplify this invariant. Return it sorted...
	return rational ();

STORE:	// We've successfully computed the invariant. Store it in the database.
	GWdbm.reset_key ();
	GWdbm.reset_value ();
	set_key (GWdbm, iv);
	set_value (GWdbm, result);
	GWdbm.write ();
	return result;
}

