/*
 * linsys.cc - a class for solving linear systems of equations in Q
 *
 */

#include <assert.h>

#include "linsys.h"

/*
 * If we include a file defining LINSYS_DEBUG, debugging output is printed
 * on every send_row function call.
 *
 */

#include "config.h"



int linsys::add_var (int v)
{
	int i,j;

	// Check if the variable is already present
	for (i=0; i<cols; i++) if (varnames[i]==v) return i;

	// Create the variable
	if ( (i = cols++) >= maxsize ) return -1;
	varnames[i] = v;
	for (j=0; j<=rows; j++) entry (j,i) = 0;
	return i;
}



void linsys::write_current_eqn ()
{
	int j;

	for (j=0; j<cols; j++) {
		if (j && (entry(rows,j)>=0)) cout << "+";
		cout << entry(rows,j) << "x[" << varnames[j] << "]";
	}
	cout << " = " << entry(rows,maxsize) << endl;
}



ostream& operator<< (ostream& os, const linsys& Q)
{
	int i,j;

	os << "system has " << Q.rows << " eqn(s) for "
	   << Q.cols << " var(s):" << endl;
	for (i=0; i<Q.rows; i++) {
		os << "  ";
		for (j=0; j<Q.cols; j++) {
			if (j && (Q.entry(i,j)>=0)) os << "+";
			os << Q.entry(i,j) << "x[" << Q.varnames[j] << "]";
		}
		os << " = " << Q.entry(i,Q.maxsize) << endl;
	}

	return os;
}



enum linsys_t linsys::send_row ()
{
	int i,j;
	rational c;

	#ifdef LINSYS_DEBUG
	++rows;
	cout << "*linsys* got new equation, " << *this;
	--rows;
	#endif

	// We now have a system of the form
	//                rows    cols   maxsize
	//                 |       |      |
	//                 V       V      V
	//         X 0 0 0 * * * *        *
	//         0 X 0 0 * * * *        *
	//         0 0 X 0 * * * *        *
	//         0 0 0 X * * * *        *
	// rows->  * * * * * * * *        *
	// where X is a non-zero and * any entry.
	//
	// Reduce new equation modulo the old ones.
	for (i=0; i<rows; i++) {
		c = entry(rows,i) / entry(i,i);

		// We must now subtract c times row i to row 'rows'.
		entry (rows,i) = 0;
		for (j=rows; j<cols; j++)
			entry (rows,j) -= c * entry (i,j);
		entry (rows,maxsize) -= c * entry (i,maxsize);
	}

	// We now have a system
	//                rows    cols   maxsize
	//                 |       |      |
	//                 V       V      V
	//         X 0 0 0 * * * *        *
	//         0 X 0 0 * * * *        *
	//         0 0 X 0 * * * *        *
	//         0 0 0 X * * * *        *
	// rows->  0 0 0 0 * * * *        *
	// Now check if there are any non-zero coefficients in the new line.
	// If so, rename variables so that the _next_ one has non-zero
	// coefficients.
	for (i=rows; i<cols; i++) if (! is_zero (entry (rows,i))) goto FOUND;
	if ( is_zero (entry (rows,maxsize)) ) {
		#ifdef LINSYS_DEBUG
		cout << "*linsys* new equation is dependent, skip"
		     << endl;
		#endif
		if (cols>rows) return UNDETERMINED;
		else return SOLVABLE;
	} else {
		#ifdef LINSYS_DEBUG
		cout << "*linsys* new equation gives a contradiction"
		     << endl;
		#endif
		return CONTRADICTION;
	}
FOUND:

	if (i!=rows) {
		// Exchange variables rows and i.
		j = varnames[i];
		varnames[i] = varnames[rows];
		varnames[rows] = j;
		for (j=0; j<=rows; j++) {
			c = entry (j,i);
			entry (j,i) = entry (j,rows);
			entry (j,rows) = c;
		}
	}

	// We now have a system
	//                rows    cols   maxsize
	//                 |       |      |
	//                 V       V      V
	//         X 0 0 0 * * * *        *
	//         0 X 0 0 * * * *        *
	//         0 0 X 0 * * * *        *
	//         0 0 0 X * * * *        *
	// rows->  0 0 0 0 X * * *        *
	// Now make the entries above the new X to 0.
	for (i=0; i<rows; i++) {
		c = entry (i,rows) / entry (rows,rows);
		entry (i,rows) = 0;
		for (j=rows+1; j<cols; j++)
			entry (i,j) -= c * entry (rows,j);
		entry (i,maxsize) -= c * entry (rows,maxsize);
	}

	// We now have a system
	//                rows    cols   maxsize
	//                 |       |      |
	//                 V       V      V
	//         X 0 0 0 0 * * *        *
	//         0 X 0 0 0 * * *        *
	//         0 0 X 0 0 * * *        *
	//         0 0 0 X 0 * * *        *
	// rows->  0 0 0 0 X * * *        *
	// Initialize a new row.
	++rows;
	for (i=0; i<cols; i++) entry (rows,i) = 0;
	entry (rows,maxsize) = 0;

	#ifdef LINSYS_DEBUG
	cout << "*linsys* new equation is independent, " << *this;
	if (cols==rows) cout << "*linsys* the system is solvable now" << endl;
	#endif

	if (cols>rows) return UNDETERMINED;
	return SOLVABLE;
}



rational linsys::value_by_number (int i)
{
	if (cols>rows) return rational();
	return entry (i,maxsize) / entry (i,i);
}



rational linsys::value_by_name (int i)
{
	int j;

	if (cols>rows) return rational ();
	for (j=0; j<cols; j++) if (varnames[j]==i)
		return entry (j,maxsize) / entry (j,j);

	return rational ();
}



rational linsys::det ()
{
	if (cols!=rows) return rational ();

	rational result = 1;
	for (int i=rows-1; i>=0; i--) result *= entry(i,i);
	return result;
}

